from functools import cache, lru_cache

from ftmq.model.mixins import YamlMixin
from pydantic import BaseModel

from juditha.cache import Cache, get_cache
from juditha.settings import FUZZY, JUDITHA, JUDITHA_CONFIG
from juditha.source import Source


class Store(BaseModel, YamlMixin):
    sources: list[Source] = []
    cache: Cache | None = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        self.cache = get_cache()

    def lookup(self, value: str, fuzzy: bool | None = FUZZY) -> str | None:
        res = self.cache.get(value)
        if res is not None:
            return res
        for source in self.sources:
            res = source.lookup(value)
            if res is not None:
                return self.cache.set(value)
        if fuzzy:
            return self.cache.fuzzy(value)

    def add(self, value: str, fuzzy: bool | None = FUZZY) -> None:
        self.cache.set(value)
        if fuzzy:
            self.cache.index(value)


@cache
def get_store(uri: str | None = None, juditha_url: str | None = None) -> Store:
    uri = uri or JUDITHA_CONFIG
    if uri:
        return Store.from_path(uri)
    url = juditha_url or JUDITHA
    if url:
        return Store(sources=[Source(klass="juditha", config={"url": url})])
    return Store()


@lru_cache(100_000)
def lookup(value: str, fuzzy: bool | None = FUZZY) -> str | None:
    store = get_store()
    return store.lookup(value, fuzzy=fuzzy)



from fastapi import FastAPI, Response
from fastapi.exceptions import HTTPException

from juditha import __version__, lookup, settings

app = FastAPI(
    debug=settings.DEBUG,
    title=settings.TITLE,
    contact=settings.CONTACT,
    description=settings.DESCRIPTION,
    version=__version__,
    redoc_url="/",
)


@app.get("/{q}")
async def api_lookup(q: str, fuzzy: bool | None = False) -> str:
    value = lookup(q, fuzzy=fuzzy)
    if value is None:
        raise HTTPException(404)
    return Response(value)


@app.head("/{q}")
async def api_head(q: str, fuzzy: bool | None = False) -> None:
    value = lookup(q, fuzzy=fuzzy)
    if value is None:
        raise HTTPException(404)
    return 200



from typing import Any, Literal, TypeAlias

from pydantic import BaseModel

from juditha.sources.aleph import Aleph
from juditha.sources.juditha import Juditha
from juditha.sources.wikipedia import Wikipedia

Stores = {
    "juditha": Juditha,
    "aleph": Aleph,
    "wikipedia": Wikipedia,
}
Sources: TypeAlias = Literal[tuple(Stores.keys())]


class Source(BaseModel):
    klass: Sources
    name: str | None = None
    config: dict[str, Any]
    store: Aleph | Wikipedia = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        self.name = self.name or f"{self.klass}-{hash(repr(self.config))}"
        self.store = Stores[self.klass].from_source(self)

    def lookup(self, value: str) -> str | None:
        return self.store.lookup(value)



"""
Use a juditha api endpoint as a source ;)
"""

from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel, HttpUrl

if TYPE_CHECKING:
    from juditha.source import Source


class Juditha(BaseModel):
    url: HttpUrl

    def lookup(self, value: str) -> str | None:
        res = requests.get(f"{self.url}/{value}")
        if res.ok:
            return res.content.decode().strip()
        return None

    @staticmethod
    def from_source(source: "Source") -> "Juditha":
        return Juditha(**source.config)



from typing import Generator

from followthemoney.types import registry
from ftmq.types import CE
from normality import normalize


def canonize(value: str) -> str:
    return normalize(value)


def names(proxy: CE) -> Generator[str, None, None]:
    for name in proxy.get_type_values(registry.name):
        yield name


def canonized_names(proxy: CE) -> Generator[str, None, None]:
    for name in names(proxy):
        yield canonize(name)


def test_proxy(proxy: CE, name: str) -> str | None:
    cname = canonize(name)
    if proxy.caption == cname:
        return proxy.caption.ljust()
    for n in names(proxy):
        if canonize(n) == cname:
            return n.lower()
    return



from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel, HttpUrl

from juditha.util import canonize

if TYPE_CHECKING:
    from juditha.source import Source


class Wikipedia(BaseModel):
    url: HttpUrl
    api_url: HttpUrl

    def __init__(self, **data):
        if "api_url" not in data:
            data["api_url"] = data["url"] + "/w/api.php"
        super().__init__(**data)

    def lookup(self, value: str) -> str | None:
        value = canonize(value)
        query = {
            "action": "opensearch",
            "format": "json",
            "formatversion": 2,
            "search": value,
            "limit": 10,
        }
        res = requests.get(self.api_url, params=query)
        assert res.ok
        res = res.json()
        for name in res[1]:
            if canonize(name) == value:
                return name.lower()

    @staticmethod
    def from_source(source: "Source") -> "Wikipedia":
        return Wikipedia(**source.config)



import os
from typing import Any

from banal import as_bool


def get_env(env: str, default: Any | None = None) -> Any | None:
    return os.environ.get(env, default)


DEBUG = as_bool(get_env("DEBUG", 0))
REDIS_URL = get_env("REDIS_URL", "redis://localhost:6379")
REDIS_PREFIX = get_env("REDIS_PREFIX", "juditha")

JUDITHA_CONFIG = get_env("JUDITHA_CONFIG")
JUDITHA = get_env("JUDITHA")
FUZZY = as_bool(get_env("FUZZY", 0))
FUZZY_SCORE = float(get_env("FUZZY_SCORE", 0.94))

# Api documentation render
TITLE = os.environ.get("TITLE", "Juditha")
CONTACT = {
    "name": os.environ.get("CONTACT_AUTHOR", "Simon Wörpel"),
    "url": os.environ.get(
        "CONTACT_URL", "https://github.com/investigativedata/juditha/"
    ),
    "email": os.environ.get("CONTACT_EMAIL", "simon@investigativedata.org"),
}
DESCRIPTION = """
Super fast canonical name lookup. Just do head requests:

    curl -I "http://localhost:8000/Berlin"
    HTTP/1.1 200 OK

    curl -I "http://localhost:8000/Bayern"
    HTTP/1.1 404 Not Found

To get the actual canonized value, do a GET request:

    curl "http://localhost:8000/berlin"
    "Berlin"
"""



"""
build an inverted token-based redis index for fuzzy lookups
"""

from typing import Set

import Levenshtein
from fingerprints import generate as fp

from juditha import settings
from juditha.util import canonize


def tokenize(value: str) -> set[str]:
    value = fp(value)
    return {t for t in value.split() if len(t) > 3}


def compare(value1: str, value2: str) -> int:
    return Levenshtein.ratio(canonize(value1), canonize(value2))


def find_best(
    value: str, candidates: Set[str], threshold: int | None = settings.FUZZY_SCORE
) -> str | None:
    if not candidates:
        return
    scores: dict[str, int] = {compare(value, c): c for c in candidates}
    score = max(scores.keys())
    if score > threshold:
        return scores[score]



import logging

from ftmq.io import smart_read, smart_read_proxies
from ftmq.model.dataset import Catalog, Dataset
from pantomime.types import FTM

from juditha.store import get_store
from juditha.util import names

log = logging.getLogger(__name__)


def _load_dataset(dataset: Dataset) -> int:
    store = get_store()
    ix = 0
    names_seen = False
    for resource in dataset.resources:
        if resource.name == "names.txt":
            names_seen = True
            for name in smart_read(resource.url, stream=True):
                store.add(name)
                ix += 1
                if ix % 10_000 == 0:
                    log.info(f"[{dataset.name}] Loading name %d ..." % ix)
    if not names_seen:
        # use entities
        for resource in dataset.resources:
            if resource.mime_type == FTM:
                for proxy in smart_read_proxies(resource.url):
                    for name in names(proxy):
                        store.add(name)
                        ix += 1
                        if ix % 10_000 == 0:
                            log.info(f"[{dataset.name}] Loading name %d ..." % ix)

    return ix


def load_dataset(uri: str) -> int:
    dataset = Dataset.from_uri(uri)
    log.info(f"[{dataset.name}] Loading ...")
    return _load_dataset(dataset)


def load_catalog(uri: str) -> int:
    catalog = Catalog.from_uri(uri)
    ix = 0
    for dataset in catalog.datasets:
        ix += _load_dataset(dataset)
    return ix


def load_names(uri: str) -> int:
    store = get_store()
    for ix, name in enumerate(smart_read(uri, stream=True), 1):
        store.add(name)
        if ix % 10_000 == 0:
            log.info("Loading name %d ..." % ix)
    return ix


def load_proxies(uri: str) -> int:
    store = get_store()
    ix = 0
    for proxy in smart_read_proxies(uri):
        for name in names(proxy):
            store.add(name)
            ix += 1
            if ix % 10_000 == 0:
                log.info("Loading name %d ..." % ix)



from typing import TYPE_CHECKING

from alephclient.api import AlephAPI
from ftmq.util import make_proxy
from pydantic import BaseModel

from juditha.util import test_proxy

if TYPE_CHECKING:
    from juditha.source import Source


class Aleph(BaseModel):
    host: str
    api_key: str | None = None
    api: AlephAPI

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data) -> "Aleph":
        data["api"] = AlephAPI(data.get("host"), data.get("api_key"))
        super().__init__(**data)

    def lookup(self, value: str) -> str | None:
        for result in self.api.search(value):
            proxy = make_proxy(result)
            name = test_proxy(proxy, value)
            if name is not None:
                return name

    @staticmethod
    def from_source(source: "Source") -> "Aleph":
        return Aleph(**source.config)



import logging
from functools import cache
from typing import Set

import fakeredis
import redis
from fingerprints import generate as fp
from normality import normalize

from juditha import settings
from juditha.index import find_best, tokenize
from juditha.util import canonize

log = logging.getLogger(__name__)


class Cache:
    def __init__(self):
        if settings.DEBUG:
            con = fakeredis.FakeStrictRedis()
            con.ping()
            log.info("Redis connected: `fakeredis`")
        else:
            con = redis.from_url(settings.REDIS_URL)
            con.ping()
            log.info(f"Redis connected: `{settings.REDIS_URL}`")
        self.cache = con

    def set(self, value: str) -> str:
        value = canonize(value)
        self.cache.set(self.get_key(value), value)
        return value

    def get(self, key: str) -> str | None:
        value = self.cache.get(self.get_key(key))
        if value is not None:
            return value.decode().strip()

    def index(self, value: str) -> int:
        value = canonize(value)
        # store fingerprint
        self.cache.set(self.get_key(fp(value)), value)
        # store inverted tokens
        tokens = tokenize(value)
        ix = 1
        for token in tokens:
            ix += self.cache.sadd(self.get_key(token) + "#SET", value)
        return ix

    def fuzzy(
        self, value: str, threshold: int | None = settings.FUZZY_SCORE
    ) -> str | None:
        value = canonize(value)
        res = self.get(fp(value))
        if res is not None:
            return res
        for token in tokenize(value):
            match = find_best(value, self.smembers(token), threshold=threshold)
            if match:
                return match

    def smembers(self, key: str) -> Set[str]:
        key = self.get_key(key) + "#SET"
        res: Set[bytes] = self.cache.smembers(key)
        return {v.decode() for v in res}

    @staticmethod
    def get_key(key: str) -> str:
        return f"{settings.REDIS_PREFIX}:{normalize(key)}"


@cache
def get_cache() -> Cache:
    return Cache()



from juditha.store import get_store, lookup


class Juditha:
    def __init__(self, url: str) -> "Juditha":
        self.store = get_store(juditha_url=url)

    def lookup(self, value: str) -> str | None:
        return self.store.lookup(value)


__version__ = "0.0.3"
__all__ = ["lookup"]



import logging
from typing import Annotated

import typer
from rich import print

from juditha import io
from juditha.store import lookup

logging.basicConfig(level=logging.INFO)

cli = typer.Typer()


def success(ix: int) -> None:
    print(f"[green]Imported {ix} names")


@cli.command("import")
def cli_import(
    uri: Annotated[str, typer.Option("-i", help="Input uri, default stdin")] = "-",
    from_entities: Annotated[
        bool, typer.Option(help="Specify if import data is ftm data (json lines)")
    ] = False,
):
    if from_entities:
        success(io.load_proxies(uri))
    else:
        success(io.load_names(uri))


@cli.command()
def load_dataset(uri: str) -> int:
    success(io.load_dataset(uri))


@cli.command()
def load_catalog(uri: str) -> int:
    success(io.load_catalog(uri))


# @cli.callback(invoke_without_command=True)
@cli.command("lookup")
def cli_lookup(value: str):
    result = lookup(value)
    if result is not None:
        print(result)
    else:
        print("[red]not found[/red]")
